import argparse
import warnings
import igraph as ig
import otf2
from itertools import chain, count, groupby
from collections import Counter
from otf2.events import Enter, Leave
from otter.trace import AttributeLookup, RegionLookup, yield_chunks, process_chunk
from otter.styling import colormap_region_type, colormap_edge_type, shapemap_region_type
from otter.helpers import set_tuples, reject_task_create, attr_handler, label_clusters, descendants_if, attr_getter, pass_master_event


def main():
    parser = argparse.ArgumentParser(
        prog="python3 -m otter",
        description='Convert an Otter OTF2 trace archive to its execution graph representation',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument('anchorfile', help='OTF2 anchor file')
    parser.add_argument('-o', '--output', dest='output', help='output file')
    parser.add_argument('-v', '--verbose', action='store_true', dest='verbose',
                        help='print chunks as they are generated')
    parser.add_argument('-i', '--interact', action='store_true', dest='interact',
                        help='drop to an interactive shell upon completion')
    parser.add_argument('-ns', '--no-style', action='store_true', default=False, dest='nostyle',
                        help='do not apply any styling to the graph nodes')
    args = parser.parse_args()

    if args.output is None and not args.interact:
        parser.error("must select at least one of -[o|i]")

    if args.interact:
        print("Otter launched interactively")

    anchorfile = args.anchorfile

    # Convert event stream into graph chunks
    print(f"loading OTF2 anchor file: {anchorfile}")
    print("generating chunks from event stream...")
    with otf2.reader.open(anchorfile) as tr:
        attr = AttributeLookup(tr.definitions.attributes)
        regions = RegionLookup(tr.definitions.regions)
        results = (process_chunk(chunk, verbose=args.verbose) for chunk in yield_chunks(tr))
        items = zip(*results)
        chunk_types = next(items)
        chain_sort_next = lambda x: sorted(chain(*next(x)), key=lambda t: t[0])
        task_links, task_crt_ts, task_leave_ts = (chain_sort_next(items) for _ in range(3))
        g_list = next(items)

    # Make function for looking up event attributes
    event_attr = attr_getter(attr)

    task_types, *_, task_ids = zip(*[r.name.split() for r in regions.values() if r.region_role == otf2.RegionRole.TASK])
    task_types, task_ids = (zip(*sorted(zip(task_types, map(int, task_ids)), key=lambda t: t[1])))

    # Gather last leave times per explicit task
    task_end_ts = {k: max(u[1] for u in v) for k, v in groupby(task_leave_ts, key=lambda t: t[0])}

    # Task tree showing parent-child links
    task_tree = ig.Graph(edges=task_links, directed=True)
    task_tree.vs['unique_id'] = task_ids
    task_tree.vs['crt_ts'] = [t[1] for t in task_crt_ts]
    task_tree.vs['end_ts'] = [task_end_ts[node.index] if node.index in task_end_ts else None for node in task_tree.vs]
    task_tree.vs['parent_index'] = list(chain((None,), list(zip(*sorted(task_links, key=lambda t: t[1])))[0]))
    task_tree.vs['task_type'] = task_types
    if not args.nostyle:
        task_tree.vs['style'] = 'filled'
        task_tree.vs['color'] = ['red' if v['task_type'] == 'implicit' else 'gray' for v in task_tree.vs]
    tt_layout = task_tree.layout_reingold_tilford()

    # Count chunks by type
    print("graph chunks created:")
    for k, v in Counter(chunk_types).items():
        print(f"  {k:18s} {v:8d}")

    # Collect all chunks
    print("combining chunks")
    g = ig.disjoint_union(g_list)
    num_nodes = g.vcount()

    print("{:20s} {:6d}".format("nodes created", num_nodes))

    if 'task_cluster_id' not in g.vs.attribute_names():
        g.vs['task_cluster_id'] = None

    # Collapse by parallel sequence ID
    print("contracting by parallel sequence ID")
    g.vs['cluster'] = label_clusters(g.vs, lambda v: v['parallel_sequence_id'] is not None, 'parallel_sequence_id')
    nodes_before = num_nodes
    g.contract_vertices(g.vs['cluster'], combine_attrs=attr_handler(attr=attr))
    num_nodes = g.vcount()
    print("{:20s} {:6d} -> {:6d} ({:6d})".format("nodes updated", nodes_before, num_nodes, num_nodes-nodes_before))

    # Collapse by single-begin/end event
    def is_single_executor(v):
        return type(v['event']) in [Enter, Leave] and event_attr(v['event'], 'region_type') == 'single_executor'
    print("contracting by single-begin/end event")
    g.vs['cluster'] = label_clusters(g.vs, is_single_executor, 'event')
    nodes_before = num_nodes
    g.contract_vertices(g.vs['cluster'], combine_attrs=attr_handler(attr=attr))
    num_nodes = g.vcount()
    print("{:20s} {:6d} -> {:6d} ({:6d})".format("nodes updated", nodes_before, num_nodes, num_nodes-nodes_before))

    # Collapse by master-begin/end event
    def is_master(v):
        return type(v['event']) in [Enter, Leave] and event_attr(v['event'], 'region_type') == 'master'
    print("contracting by master-begin/end event")
    g.vs['cluster'] = label_clusters(g.vs, is_master, 'event')
    nodes_before = num_nodes
    g.contract_vertices(g.vs['cluster'], combine_attrs=attr_handler(events=pass_master_event, attr=attr))
    num_nodes = g.vcount()
    print("{:20s} {:6d} -> {:6d} ({:6d})".format("nodes updated", nodes_before, num_nodes, num_nodes-nodes_before))

    # Itermediate clean-up: for each master region, remove edges that connect 
    # the same nodes as the master region
    master_enter = filter(lambda v: event_attr(v['event'], 'region_type') == 'master' and event_attr(v['event'], 'endpoint') == 'enter', g.vs)
    master_enter_nodes = {v['event']: v for v in master_enter}
    master_leave = filter(lambda v: event_attr(v['event'], 'region_type') == 'master' and event_attr(v['event'], 'endpoint') == 'leave', g.vs)
    master_node_pairs = ((master_enter_nodes[leave_node['master_enter_event']], leave_node) for leave_node in master_leave)
    def yield_neighbours():
        for enter_node, leave_node in master_node_pairs:
            (p,), (s,) = enter_node.predecessors(), leave_node.successors()
            yield p, s
    neighbour_set = {(p,s) for p, s in yield_neighbours()}
    redundant_edges = list(filter(lambda e: (e.source_vertex, e.target_vertex) in neighbour_set, g.es))
    print(f"deleting redundant edges due to master regions: {len(redundant_edges)}")
    g.delete_edges(redundant_edges)

    # Collapse by (task-ID, endpoint) to get 1 subgraph per task
    for v in g.vs:
        if type(v['event']) in [Enter, Leave] and event_attr(v['event'], 'region_type') == 'explicit_task':
            v['task_cluster_id'] = (event_attr(v['event'], 'unique_id'), event_attr(v['event'], 'endpoint'))
    print("contracting by task ID & endpoint")
    g.vs['cluster'] = label_clusters(g.vs, lambda v: v['task_cluster_id'] is not None, 'task_cluster_id')
    nodes_before = num_nodes
    g.contract_vertices(g.vs['cluster'],
                        combine_attrs=attr_handler(events=reject_task_create, tuples=set_tuples, attr=attr))
    num_nodes = g.vcount()
    print("{:20s} {:6d} -> {:6d} ({:6d})".format("nodes updated", nodes_before, num_nodes, num_nodes-nodes_before))

    # Collapse by task ID where there are no links between to combine task nodes with nothing nested within
    def is_empty_task_region(v):
        if v['task_cluster_id'] is None:
            return False
        if type(v['event']) in [Enter, Leave]:
            return (type(v['event']) is Leave and v.indegree() == 0) or \
                   (type(v['event']) is Enter and v.outdegree() == 0)
        if type(v['event']) is list and set(map(type, v['event'])) in [{Enter}, {Leave}]:
            return (set(map(type, v['event'])) == {Leave} and v.indegree() == 0) or \
                   (set(map(type, v['event'])) == {Enter} and v.outdegree() == 0)
    print("contracting by task ID where there are no nested nodes")
    g.vs['cluster'] = label_clusters(g.vs, is_empty_task_region, lambda v: v['task_cluster_id'][0])
    nodes_before = num_nodes
    g.contract_vertices(g.vs['cluster'], combine_attrs=attr_handler(events=reject_task_create, tuples=set_tuples, attr=attr))
    num_nodes = g.vcount()
    print("{:20s} {:6d} -> {:6d} ({:6d})".format("nodes updated", nodes_before, num_nodes, num_nodes-nodes_before))

    # Collapse redundant sync-enter/leave node pairs by labelling unique pairs of nodes identified by their shared edge
    dummy_counter = count()
    for e in g.es:
        node_types = set()
        for v in (e.source_vertex, e.target_vertex):
            if type(v['event']) is not list:
                node_types.add(event_attr(v['event'], 'region_type'))
            else:
                for event in v['event']:
                    node_types.add(event_attr(event, 'region_type'))
        if node_types in [{'barrier_implicit'}, {'barrier_explicit'}, {'taskwait'}, {'loop'}] and \
                e.source_vertex.attributes().get('sync_cluster_id', None) is None and \
                e.target_vertex.attributes().get('sync_cluster_id', None) is None:
            value = next(dummy_counter)
            e.source_vertex['sync_cluster_id'] = e.target_vertex['sync_cluster_id'] = value
    print("contracting redundant sync-enter/leave node pairs")
    g.vs['cluster'] = label_clusters(g.vs, lambda v: v['sync_cluster_id'] is not None, 'sync_cluster_id')
    nodes_before = num_nodes
    g.contract_vertices(g.vs['cluster'], combine_attrs=attr_handler(tuples=set_tuples, attr=attr))
    num_nodes = g.vcount()
    print("{:20s} {:6d} -> {:6d} ({:6d})".format("nodes updated", nodes_before, num_nodes, num_nodes-nodes_before))

    # Unpack the region_type attribute
    for v in g.vs:
        if type(v['event']) is list:
            v['region_type'], = set([event_attr(e, 'region_type') for e in v['event']])
            v['endpoint'] = set([event_attr(e, 'endpoint') for e in v['event']])
        else:
            v['region_type'] = event_attr(v['event'], 'region_type')
            v['endpoint'] = event_attr(v['event'], 'endpoint')
        if type(v['endpoint']) is set and len(v['endpoint']) == 1:
            v['endpoint'], = v['endpoint']

    # Apply taskwait synchronisation
    print("applying taskwait synchronisation")
    for twnode in g.vs.select(lambda v: v['region_type'] == 'taskwait'):
        parents = set(task_tree.vs[event_attr(e, 'encountering_task_id')] for e in twnode['event'])
        tw_encounter_ts = {event_attr(e, 'encountering_task_id'): e.time for e in twnode['event'] if type(e) is Enter}
        children = [c.index for c in chain(*[p.neighbors(mode='out') for p in parents])
                    if c['crt_ts'] < tw_encounter_ts[c['parent_index']] < c['end_ts']]
        nodes = [v for v in g.vs if v['region_type'] == 'explicit_task'
                                 and event_attr(v['event'], 'unique_id') in children
                                 and v['endpoint'] != 'enter']
        ecount = g.ecount()
        g.add_edges([(v.index, twnode.index) for v in nodes])
        g.es[ecount:]['type'] = 'taskwait'

    def event_time_per_task(event):
        """Return the map: encountering task id -> event time for all encountering tasks in the event"""
        if type(event) is list:
            return {event_attr(e, 'encountering_task_id'): e.time for e in event}
        return {event_attr(event, 'encountering_task_id'): event.time}

    # Apply taskgroup synchronisation
    print("applying taskgroup synchronisation")
    for tgnode in g.vs.select(lambda v: v['region_type'] == 'taskgroup' and v['endpoint'] == 'leave'):
        tg_enter_ts = event_time_per_task(tgnode['taskgroup_enter_event'])
        tg_leave_ts = event_time_per_task(tgnode['event'])
        parents = [task_tree.vs[k] for k in tg_enter_ts]
        children = [c for c in chain(*[p.neighbors(mode='out') for p in parents])
                    if tg_enter_ts[c['parent_index']] < c['crt_ts'] < tg_leave_ts[c['parent_index']]]
        descendants = list(chain(*[descendants_if(c, cond=lambda x: x['task_type'] != 'implicit') for c in children]))
        nodes = [v for v in g.vs if v['region_type'] == 'explicit_task'
                                 and event_attr(v['event'], 'unique_id') in descendants
                                 and v['endpoint'] != 'enter']
        ecount = g.ecount()
        g.add_edges([(v.index, tgnode.index) for v in nodes])
        g.es[ecount:]['type'] = 'taskgroup'

    # Apply styling if desired
    if not args.nostyle:
        print("applying node and edge styline")
        g.vs['color'] = [colormap_region_type[v['region_type']] for v in g.vs]
        g.vs['style'] = 'filled'
        g.vs['shape'] = [shapemap_region_type[v['region_type']] for v in g.vs]
        g.es['color'] = [colormap_edge_type[e.attributes().get('type', None)] for e in g.es]
    g.vs['label'] = ["{}".format(event_attr(v['event'], 'unique_id'))
                     if any(s in v['region_type'] for s in ['explicit', 'initial', 'parallel']) else " " for v in g.vs]

    g.simplify(combine_edges='first')

    # Clean up redundant attributes
    for item in ['task_cluster_id', 'parallel_sequence_id', 'cluster', 'sync_cluster_id']:
        if item in g.vs.attribute_names():
            print(f"deleting vertex attribute '{item}'")
            del g.vs[item]

    if args.output:
        print(f"writing graph to '{args.output}'")
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            try:
                g.write(args.output)
            except OSError as oserr:
                print(f"igraph error: {oserr}")
                print(f"failed to write to file '{args.output}'")

    if args.interact:
        import atexit
        import code
        import os
        import readline
        readline.parse_and_bind("tab: complete")

        hfile = os.path.join(os.path.expanduser("~"), ".otter_history")

        try:
            readline.read_history_file(hfile)
            numlines = readline.get_current_history_length()
        except FileNotFoundError:
            open(hfile, 'wb').close()
            numlines = 0

        def append_history(n, f):
            newlines = readline.get_current_history_length()
            readline.set_history_length(1000)
            readline.append_history_file(newlines - n, f)

        atexit.register(append_history, numlines, hfile)

        k = ""
        for k, v in locals().items():
            if g is v:
                break

        banner = \
f"""
Graph '{k}' has {g.vcount()} nodes and {g.ecount()} edges

Entering interactive mode, use:
    ig.plot({k}, [target="..."], ...)   to view or plot to file
    {k}.write_*()                       to save a representation of the graph e.g. {k}.write_dot("graph.dot")     
"""
        Console = code.InteractiveConsole(locals=locals())
        Console.interact(banner=banner, exitmsg=f"history saved to {hfile}")


if __name__ == "__main__":
    main()



import re
import typing as T
import igraph as ig
import otf2
from itertools import chain
from collections import defaultdict, deque
from otf2.events import Enter, Leave, ThreadTaskCreate
from otter.helpers import attr_getter, events_bridge_region

class DefinitionLookup:

    def __init__(self, registry: otf2.registry._RefRegistry):
        self._lookup = dict()
        for d in registry:
            if d.name in self._lookup:
                raise KeyError("{} already present".format(d.name))
            self._lookup[d.name] = d

    def __getitem__(self, name):
        if name in self._lookup:
            return self._lookup[name]
        else:
            raise AttributeError(name)

    def __iter__(self):
        return ((k,v) for k, v in self._lookup.items())

    def keys(self):
        return self._lookup.keys()

    def values(self):
        return self._lookup.values()

    def items(self):
        return self.__iter__()


class AttributeLookup(DefinitionLookup):

    def __init__(self, attributes: otf2.registry._RefRegistry):
        if not isinstance(attributes[0], otf2.definitions.Attribute):
            raise TypeError(type(attributes[0]))
        super().__init__(attributes)

    def __repr__(self):
        s = "{:24s} {:12s} {}\n".format("Name", "Type", "Description")
        format = lambda k, v: "{:24s} {:12s} {}".format(k, str(v.type).split(".")[1], v.description)
        return s+"\n".join([format(k, v) for k,v in self._lookup.items()])


class LocationLookup(DefinitionLookup):

    def __init__(self, locations: otf2.registry._RefRegistry):
        if not isinstance(locations[0], otf2.definitions.Location):
            raise TypeError(type(locations[0]))
        super().__init__(locations)

    def __repr__(self):
        s = "{:12s} {:12s} {:12s} {}\n".format("Group", "Name", "Type", "Events")
        format = lambda v: "{:12s} {:12s} {:12s} {}".format(v.group.name, v.name, str(v.type).split(".")[1], v.number_of_events)
        return s+"\n".join([format(v) for k,v in self._lookup.items()])


class RegionLookup(DefinitionLookup):

    def __init__(self, regions: otf2.registry._RefRegistry):
        self._lookup = dict()
        for r in regions:
            ref = int(re.search(r'\d+', repr(r))[0])
            if ref in self._lookup:
                raise KeyError("{} already present".format(ref))
            self._lookup[ref] = r

    def __repr__(self):
        minref, maxref = min(self._lookup.keys()), max(self._lookup.keys())
        s = "{:3s} {:18s} {}\n".format("Ref", "Name", "Role")
        format_item = lambda l, k: "{:3d} {:18s} {}".format(k, l[k].name, str(l[k].region_role).split(".")[1])
        return s + "\n".join([format_item(self._lookup, i) for i in range(minref, maxref+1)])


class LocationEventMap:
    """
    Behaves like a collection of dicts, each of which maps a location (thread) onto some sequence of events recorded
    by that location
    """

    def __init__(self, events: T.Iterable, attr: AttributeLookup):
        self._map = defaultdict(deque)
        self.attr = attr
        for l, e in events:
            self._map[l].append(e)

    def __repr__(self):
        s = ""
        for l, q in self._map.items():
            s += l.name + "\n"
            s += "  {:18s} {:10s} {:20s} {:20s} {:18s} {}\n".format("Time", "Endpoint", "Region Type", "Event Type", "Region ID/Name", "CPU")
            for e in q:
                s += "  {:<18d} {:10s} {:20s} {:20s} {:18s} {}\n".format(
                    e.time,
                    e.attributes[self.attr['endpoint']],
                    e.attributes.get(self.attr['region_type'], ""),
                    e.attributes[self.attr['event_type']],
                    str(e.attributes[self.attr['unique_id']]) if self.attr['unique_id'] in e.attributes else e.region.name,
                    e.attributes[self.attr['cpu']])
        return s

    def __getitem__(self, location):
        if location in self._map:
            return self._map[location]
        else:
            raise KeyError(location)

    def locations(self):
        return sorted(self._map.keys(), key=lambda q: int(q.name.split()[1])) # "Thread x"

    def items(self):
        return ((l, self[l]) for l in self.locations())

    def append(self, l, e):
        self._map[l].append(e)

    @property
    def kind(self):
        _, (event, *ignore) = next(self.items())
        return event.attributes[self.attr['region_type']]


def yield_chunks(tr):
    attr = AttributeLookup(tr.definitions.attributes)
    lmap_dict = defaultdict(lambda : LocationEventMap(list(), attr))
    stack_dict = defaultdict(deque)
    nChunks = 0
    print("yielding chunks:", end=" ", flush=True)
    for location, event in tr.events:
        if type(event) in [otf2.events.ThreadBegin, otf2.events.ThreadEnd]:
            continue
        if event_defines_new_chunk(event, attr):
            # Event marks transition from one chunk to another
            if isinstance(event, Enter):
                if event.attributes.get(attr['region_type'], "") != 'explicit_task':
                    lmap_dict[location].append(location, event)
                stack_dict[location].append(lmap_dict[location])
                # New location map for new chunk
                lmap_dict[location] = LocationEventMap([(location, event)], attr)
            elif isinstance(event, Leave):
                lmap_dict[location].append(location, event)
                nChunks += 1
                if nChunks % 100 == 0:
                    if nChunks % 1000 == 0:
                        print("yielding chunks:", end=" ", flush=True)
                    print(f"{nChunks:4d}", end="", flush=True)
                elif nChunks % 20 == 0:
                    print(".", end="", flush=True)
                if (nChunks+1) % 1000 == 0:
                    print("", flush=True)
                yield lmap_dict[location]
                # Continue with enclosing chunk
                lmap_dict[location] = stack_dict[location].pop()
                if event.attributes.get(attr['region_type'], "") != 'explicit_task':
                    lmap_dict[location].append(location, event)
            else:
                lmap_dict[location].append(location, event)
        else:
            # Append event to current chunk for this location
            lmap_dict[location].append(location, event)
    print("")


def event_defines_new_chunk(e: otf2.events._EventMeta, a: AttributeLookup) -> bool:
    return (e.attributes.get(a['region_type'], None) in ['parallel', 
        'explicit_task', 'initial_task', 'single_executor', 'master'])


def process_chunk(chunk, verbose=False):
    """Return a tuple of chunk kind, task-create links, task-create times, task-leave times and the chunk's graph"""

    # Make function for looking up event attributes
    get_attr = attr_getter(chunk.attr)

    # Unpack events from chunk
    (_, (first_event, *events, last_event)), = chunk.items()

    if verbose and len(events) > 0:
        print(chunk)

    # Make the graph representing this chunk
    g = ig.Graph(directed=True)
    prior_node = g.add_vertex(event=first_event)

    # Used to save taskgroup-enter event to match to taskgroup-leave event
    taskgroup_enter_event = None

    # Match master-enter event to corresponding master-leave
    master_enter_event = first_event if get_attr(first_event, 'region_type') == 'master' else None

    if chunk.kind == 'parallel':
        parallel_id = get_attr(first_event, 'unique_id')
        prior_node["parallel_sequence_id"] = (parallel_id, get_attr(first_event, 'endpoint'))

    task_create_nodes = deque()
    task_links = deque()
    task_crt_ts = deque()
    task_leave_ts = deque()

    if type(first_event) is Enter and get_attr(first_event, 'region_type') in ['initial_task']:
        task_crt_ts.append((get_attr(first_event, 'unique_id'), first_event.time))

    k = 1
    for event in chain(events, (last_event,)):

        if get_attr(event, 'region_type') in ['implicit_task']:
            if type(event) is Enter:
                task_links.append((get_attr(event, 'encountering_task_id'), get_attr(event, 'unique_id')))
                task_crt_ts.append((get_attr(event, 'unique_id'), event.time))
            elif type(event) is Leave:
                task_leave_ts.append((get_attr(event, 'unique_id'), event.time))
            continue

        # The node representing this event
        node = g.add_vertex(event=event)

        # Add task-leave time
        if type(event) is Leave and get_attr(event, 'region_type') == 'explicit_task':
            task_leave_ts.append((get_attr(event, 'unique_id'), event.time))

        # Add task links and task crt ts
        if (type(event) is Enter and get_attr(event, 'region_type') == 'implicit_task') \
                or (type(event) is ThreadTaskCreate):
            task_links.append((get_attr(event, 'encountering_task_id'), get_attr(event, 'unique_id')))
            task_crt_ts.append((get_attr(event, 'unique_id'), event.time))

        # Match taskgroup-enter/-leave events
        if get_attr(event, 'region_type') in ['taskgroup']:
            if type(event) is Enter:
                taskgroup_enter_event = event
            elif type(event) is Leave:
                if taskgroup_enter_event is None:
                    raise ValueError("taskgroup-enter event was None")
                node['taskgroup_enter_event'] = taskgroup_enter_event
                taskgroup_enter_event = None

        # Match master-enter/-leave events
        if get_attr(event, 'region_type') in ['master']:
            if type(event) is Enter:
                master_enter_event = event
            elif type(event) is Leave:
                if master_enter_event is None:
                    raise ValueError("master-enter event was None")
                node['master_enter_event'] = master_enter_event
                master_enter_event = None

        # Label nodes in a parallel chunk by their position for easier merging
        if (chunk.kind == 'parallel' 
                and type(event) is not ThreadTaskCreate
                and get_attr(event, 'region_type') != 'master'):
            node["parallel_sequence_id"] = (parallel_id, k)
            k += 1

        if get_attr(event, 'region_type') == 'parallel':
            # Label nested parallel regions for easier merging...
            if event is not last_event:
                node["parallel_sequence_id"] = (get_attr(event, 'unique_id'), get_attr(event, 'endpoint'))
            # ... but distinguish from a parallel chunk's terminating parallel-end event
            else:
                node["parallel_sequence_id"] = (parallel_id, get_attr(event, 'endpoint'))

        # Add edge except for (single begin -> single end) and (parallel N begin -> parallel N end)
        if events_bridge_region(prior_node['event'], node['event'], ['single_executor', 'single_other', 'master'], get_attr) \
            or (events_bridge_region(prior_node['event'], node['event'], ['parallel'], get_attr)
                and get_attr(node['event'], 'unique_id') == get_attr(prior_node['event'], 'unique_id')):
            pass
        else:
            g.add_edge(prior_node, node)

        # For task_create add dummy nodes for easier merging
        if type(event) is ThreadTaskCreate:
            node['task_cluster_id'] = (get_attr(event, 'unique_id'), 'enter')
            dummy_node = g.add_vertex(event=event, task_cluster_id=(get_attr(event, 'unique_id'), 'leave'))
            task_create_nodes.append(dummy_node)
            continue
        elif len(task_create_nodes) > 0:
            task_create_nodes = deque()

        prior_node = node

    if chunk.kind == 'explicit_task' and len(events) == 0:
        g.delete_edges([0])

    # Require at least 1 edge between start and end nodes if there are no internal nodes, except for empty explicit
    # task chunks
    if chunk.kind != "explicit_task" and len(events) == 0 and g.ecount() == 0:
        g.add_edge(g.vs[0], g.vs[1])

    return chunk.kind, task_links, task_crt_ts, task_leave_ts, g



from collections import defaultdict

"""
Styling dictionaries for __main__.py
"""

# Map region type to node color
colormap_region_type = defaultdict(lambda: 'white', **{
    'initial_task': 'green',
    'implicit_task': 'fuchsia',
    'explicit_task': 'cyan',
    'parallel': 'yellow',
    'single_executor': 'blue',
    'single_other': 'orange',
    'taskwait': 'red',
    'taskgroup': 'purple',
    'barrier_implicit': 'darkgreen',
    'master': 'magenta',

    # Workshare regions
    'loop': 'brown',
    'taskloop': 'orange',

    # For colouring by endpoint
    'enter': 'green',
    'leave': 'red'
})

colormap_edge_type = defaultdict(lambda: 'black', **{
    'taskwait': 'red',
    'taskgroup': 'red',
})

shapemap_region_type = defaultdict(lambda: 'circle', **{
    'initial_task': 'square',
    'implicit_task': 'square',
    'explicit_task': 'square',
    'parallel': 'parallelogram',

    # Sync regions
    'taskwait': 'octagon',
    'taskgroup': 'octagon',
    'barrier_implicit': 'octagon',

    # Workshare regions
    'loop': 'diamond',
    'taskloop': 'diamond',
    'single_executor': 'diamond',

    # Master
    'master': 'circle'
})



"""
Key functionality:

- load OTF2 trace and extract locations, events, regions, attributes so they can be easily looked up
- parse loaded trace to determine what graph nodes to create and where edges are required
- export execution graph to file
"""


from itertools import chain, count
from collections import defaultdict
from otf2.events import Enter, Leave, ThreadTaskCreate

"""
Contains assorted helper functions for __main__.py and trace.py
"""


def attr_getter(attr_lookup):
    """Make a function to lookup an attribute of some event by its name"""
    def event_attr_getter(evt, name):
        if type(evt) is list:
            result, = set([e.attributes[attr_lookup[name]] for e in evt])
            return result
        elif type(evt) in [Enter, Leave, ThreadTaskCreate]:
            return evt.attributes[attr_lookup[name]]
        else:
            raise TypeError(f"unexpected type: {type(evt)}")
    return event_attr_getter


# Helpers related to manipulating the nodes of an igraph.Graph:


def chain_lists(lists):
    """Possible argument to attr_handler()"""
    return list(chain(*lists))


def set_tuples(tuples):
    """Possible argument to attr_handler()"""
    s = set(tuples)
    if len(s) == 1:
        return s.pop()
    else:
        return s


def pass_args(args):
    """Possible argument to attr_handler()"""
    return args


def pass_single_executor(events, **kw):
    """Possible argument to attr_handler()"""
    region_types = {e.attributes[kw['attr']['region_type']] for e in events}
    if region_types == {'single_other', 'single_executor'}:
        single_executor, = filter(lambda e: e.attributes[kw['attr']['region_type'] ]=='single_executor', events)
        return single_executor
    else:
        return events

def pass_master_event(events, **kw):
    """Possible argument to attr_handler()"""
    region_types = {e.attributes[kw['attr']['region_type']] for e in events}
    if region_types == {'master'} and len(set(events)) == 1:
        return events[0]
    else:
        return events


def reject_task_create(events, **kw):
    """Possible argument to attr_handler()"""
    events = [e for e in events if type(e) is not ThreadTaskCreate]
    if len(events) == 1:
        return events[0]
    else:
        return events


def attr_handler(events=pass_single_executor, ints=min, lists=chain_lists, tuples=set, **kw):
    """Make a function for combining lists of node attributes according to their type"""
    def attr_combiner(args):
        if len(args) == 1:
            return args[0]
        else:
            if all([isinstance(obj, int) for obj in args]):
                return ints(args)
            elif all([isinstance(obj, list) for obj in args]):
                return lists(args)
            elif all([isinstance(obj, tuple) for obj in args]):
                return tuples(args)
            elif all([type(obj) in [Enter, Leave, ThreadTaskCreate] for obj in args]):
                return events(args, **kw)
            else:
                return args[0]
    return attr_combiner


def label_clusters(vs, condition, key):
    """Return cluster labels (given by key function) where condition is true, or a unique vertex label otherwise"""
    if isinstance(key, str):
        s = key
        key = lambda v: v[s]
    vertex_counter = count()
    cluster_counter = count(start=sum(not condition(v) for v in vs))
    label = defaultdict(lambda: next(cluster_counter))
    return [label[key(v)] if condition(v) else next(vertex_counter) for v in vs]


def graph_slices(graph, s, t):
    """Yield vertex ids in slices of the graph generated from the mincuts from s to t"""
    mincuts = graph.all_st_mincuts(s, t)
    vids = [set(c.partition[0]) for c in mincuts]
    yield list(vids[0])
    yield from [list(set(p)-set(q)) for p, q in zip(vids[1:], vids[0:-1])]
    yield list(set(vids[-0]) - set(vids[-1]))


# Functions useful in one specific place in the code:

def descendants_if(node, cond=lambda x: True):
    """Yield all descendants D of node, skipping E & its descendants if cond(E) is False."""
    for child in node.successors():
        if cond(child):
            yield from descendants_if(child, cond=cond)
    yield node.index


def events_bridge_region(previous, current, types, getter):
    """Used in trace.process_chunk to check for certain enter-leave event sequences"""
    return (getter(previous, 'region_type') in types
            and getter(previous, 'endpoint') == 'enter'
            and getter(current, 'region_type') in types
            and getter(current, 'endpoint') == 'leave')



from setuptools import find_packages, setup
from pyotter_version import version

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(name="pyotter",
    version=".".join(version),
    author="Adam Tuft",
    author_email='adam.s.tuft@gmail.com',
    description="Otter post-processing tool",
    long_description=long_description,
    long_description_content_type="text/markdown",
    platforms=["linux"],
    license="https://github.com/adamtuft/pyotter/blob/main/LICENSE",
    url="https://github.com/adamtuft/pyotter",
    packages=find_packages(),
    install_requires=[
        'python-igraph==0.9.1'
    ],
    dependency_links=['https://perftools.pages.jsc.fz-juelich.de/cicd/otf2/']
)
